from seesaw.externalprocess import ExternalProcess
from seesaw.pipeline import Pipeline
from seesaw.runner import SimpleRunner
from seesaw.task import PrintItem, SimpleTask
from seesaw.test_base import BaseTestCase


class ExternalProcessTest(BaseTestCase):
    def test_max_items(self):
        pipeline = Pipeline(PrintItem(), PrintItem())
        pipeline.has_failed = None

        def fail_callback(task, item):
            pipeline.has_failed = True

        pipeline.on_fail_item += fail_callback

        runner = SimpleRunner(pipeline, max_items=3)

        def finish_item_callback(runner, pipeline, item):
            if runner.item_count > 10:
                raise Exception('Too many items.')

        runner.on_pipeline_finish_item += finish_item_callback
        runner.start()

        self.assertFalse(pipeline.has_failed)
        self.assertEqual(3, runner.item_count)
        self.assertIOLoopOK()

    def test_max_items_with_subproc(self):
        pipeline = Pipeline(PrintItem(), PrintItem(),
            ExternalProcess("pwd", ["pwd"]))
        pipeline.has_failed = None

        def fail_callback(task, item):
            pipeline.has_failed = True

        pipeline.on_fail_item += fail_callback

        runner = SimpleRunner(pipeline, max_items=3)

        def finish_item_callback(runner, pipeline, item):
            if runner.item_count > 10:
                raise Exception('Too many items.')

        runner.on_pipeline_finish_item += finish_item_callback
        runner.start()

        self.assertFalse(pipeline.has_failed)
        self.assertEqual(3, runner.item_count)
        self.assertIOLoopOK()

    def test_no_stack_overflow(self):
        pipeline = Pipeline(PrintItem())
        pipeline.has_failed = None

        def fail_callback(task, item):
            pipeline.has_failed = True

        pipeline.on_fail_item += fail_callback

        runner = SimpleRunner(pipeline, max_items=50)

        def finish_item_callback(runner, pipeline, item):
            if runner.item_count > 200:
                raise Exception('Too many items.')

        runner.on_pipeline_finish_item += finish_item_callback
        runner.start()

        self.assertFalse(pipeline.has_failed)
        self.assertEqual(50, runner.item_count)
        self.assertIOLoopOK()

    def test_spurious_item_events(self):
        class StupidTask(SimpleTask):
            def __init__(self):
                SimpleTask.__init__(self, "StupidTask")

            def process(self, item):
                item.log_output('Failing the item.')
                self.fail_item(item)
                item.log_output('Completing the item.')
                self.complete_item(item)
                item.log_output('Failing the item.')
                self.fail_item(item)

        pipeline = Pipeline(StupidTask())
        pipeline.fail_count_test = 0

        def fail_callback(task, item):
            pipeline.fail_count_test += 1

        pipeline.on_fail_item += fail_callback

        runner = SimpleRunner(pipeline, max_items=1)
        runner.start()

        self.assertEqual(1, pipeline.fail_count_test)
        self.assertIOLoopOK()



from __future__ import print_function

from argparse import ArgumentParser
import os.path
import re
import sys

import seesaw
from seesaw.runner import SimpleRunner
from seesaw.web import start_runner_server


seesaw.runner_type = "Standalone"


def load_pipeline(pipeline_path, context):
    dirname, dummy = os.path.split(pipeline_path)
    if dirname == "":
        dirname = "."

    with open(pipeline_path) as f:
        pipeline_str = f.read()

    local_context = context
    global_context = context
    curdir = os.getcwd()
    try:
        os.chdir(dirname)
        exec(pipeline_str, local_context, global_context)
    finally:
        os.chdir(curdir)

    project = local_context["project"]
    pipeline = local_context["pipeline"]
    pipeline.project = project
    return (project, pipeline)


def check_downloader_or_exit(value, regex="^[-_a-zA-Z0-9]{3,30}$"):
    if not re.search(regex, value):
        print('Please use a nickname containing only letters A-Z, '
              'numbers, underscore, and hyphen-minus only.')
        sys.exit(1)


def main():
    parser = ArgumentParser(description="Run the pipeline")
    parser.add_argument("pipeline", metavar="PIPELINE", type=str,
                        help="the pipeline file")
    parser.add_argument("downloader", metavar="DOWNLOADER", type=str,
                        help="your nickname")
    parser.add_argument("--concurrent", dest="concurrent_items",
                        help="work on N items at a time (default: 1)",
                        metavar="N", type=int, default=1)
    parser.add_argument("--max-items", dest="max_items",
                        help="stop after completing N items",
                        metavar="N", type=int, default=None)
    parser.add_argument("--stop-file", dest="stop_file",
                        help="the STOP file to be monitored (default: STOP)",
                        metavar="FILE", type=str, default="STOP")
    parser.add_argument("--disable-web-server", dest="enable_web_server",
                        help="disable the web interface",
                        action="store_false")
    parser.add_argument("--keep-data", dest="keep_data",
                        help="do not remove data of finished items",
                        action="store_true")
    parser.add_argument("--address", dest="address",
                        help="the IP address of the web interface "
                             "(default: 0.0.0.0)",
                        metavar="HOST", type=str, default="0.0.0.0")
    parser.add_argument("--port", dest="port_number",
                        help="the port number for the web interface "
                             "(default: 8001)",
                        metavar="PORT", type=int, default=8001)
    parser.add_argument("--http-username", dest="http_username",
                        help="username for the web interface (default: admin)",
                        metavar="USERNAME", type=str
                        )  # default is set in start_runner_server
    parser.add_argument("--http-password", dest="http_password",
                        help="password for the web interface",
                        metavar="PASSWORD", type=str)
    parser.add_argument("--context-value", dest="context_values",
                        help="additional pipeline global variables "
                             "(name=text)",
                        metavar='VALUE_PAIR',
                        action='append', default=[], type=str)
    parser.add_argument("--version", action="version",
                        version=seesaw.__version__)
    args = parser.parse_args()

    check_downloader_or_exit(args.downloader)

    context = {"downloader": args.downloader}

    for context_value in args.context_values:
        name, text_value = context_value.split("=", 1)
        if name not in context:
            context[name] = text_value
        else:
            raise Exception("Context value name %s already defined." % name)

    (project, pipeline) = load_pipeline(args.pipeline, context)

    print("*" * 74)
    print("*%-072s*" % " ")
    print("*%-072s*" % ("   ArchiveTeam Seesaw kit - %s" % seesaw.__version__))
    print("*%-072s*" % " ")
    print("*" * 74)
    print()
    print("Initializing pipeline for '%s'" % (project.title))
    print()
    print(pipeline)
    print()
    print("-" * 74)
    print()

    runner = SimpleRunner(
        pipeline,
        stop_file=args.stop_file,
        concurrent_items=args.concurrent_items,
        max_items=args.max_items,
        keep_data=args.keep_data)

    if args.enable_web_server:
        print("Starting the web interface on %s:%d..." %
              (args.address, args.port_number))
        print()
        print("-" * 74)
        print()
        start_runner_server(project, runner,
                            bind_address=args.address,
                            port_number=args.port_number,
                            http_username=args.http_username,
                            http_password=args.http_password)

    print("Run 'touch %s' to stop downloading." % args.stop_file)
    print()
    runner.start()


if __name__ == "__main__":
    main()



'''Contacting the work unit server.

A Tracker refers to the Universal Tracker
(https://github.com/ArchiveTeam/universal-tracker).
'''
import json
import functools
import datetime
import os.path
import re

from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from tornado.ioloop import IOLoop

import seesaw
from seesaw.config import realize
from seesaw.task import Task, SimpleTask
from seesaw.externalprocess import RsyncUpload, CurlUpload


class TrackerRequest(Task):
    '''Represents a request to a Tracker.'''
    def __init__(self, name, tracker_url, tracker_command,
                 may_be_canceled=False):
        Task.__init__(self, name)
        self.http_client = AsyncHTTPClient()
        self.tracker_url = tracker_url
        self.tracker_command = tracker_command
        self.retry_delay = 30
        self._set_may_be_canceled = may_be_canceled

    def enqueue(self, item):
        self.start_item(item)
        item.log_output("Starting %s for %s\n" % (self, item.description()))
        self.send_request(item)

    def send_request(self, item):
        if item.canceled:
            return

        if self._set_may_be_canceled:
            item.may_be_canceled = False
        self.http_client.fetch(
            HTTPRequest(
                "%s/%s" % (self.tracker_url, self.tracker_command),
                method="POST",
                headers={"Content-Type": "application/json"},
                user_agent=("ArchiveTeam Warrior/%s %s %s" % (
                    seesaw.__version__, seesaw.runner_type,
                    seesaw.warrior_build)).strip(),
                body=json.dumps(self.data(item))
                ),
            functools.partial(self.handle_response, item))

    def data(self, item):
        return {}

    def handle_response(self, item, response):
        if response.code == 200:
            self.process_body(response.body, item)
        else:
            if response.code == 420 or response.code == 429:
                r = ("Tracker rate limiting is active. "
                     "We don't want to overload the site we're archiving, "
                     "so we've limited the number of downloads per minute. "
                     "Please wait... ")
            elif response.code == 404:
                r = "No item received. "
            elif response.code == 455:
                r = "Project code is out of date and needs to be upgraded. "
            elif response.code == 599:
                r = "No HTTP response received from tracker. "
            else:
                r = "Tracker returned status code %d. \n" % (response.code)
            self.schedule_retry(item, r)

    def schedule_retry(self, item, message=""):
        if self._set_may_be_canceled:
            item.may_be_canceled = True
        item.log_output(
            "%sRetrying after %d seconds...\n" % (message, self.retry_delay))
        IOLoop.instance().add_timeout(
            datetime.timedelta(seconds=self.retry_delay),
            functools.partial(self.send_request, item))

    def process_body(self, body, item):
        raise NotImplementedError()


class GetItemFromTracker(TrackerRequest):
    '''Get a single work unit information from the Tracker.'''
    def __init__(self, tracker_url, downloader, version=None):
        TrackerRequest.__init__(self, "GetItemFromTracker", tracker_url,
                                "request", may_be_canceled=True)
        self.downloader = downloader
        self.version = version

    def data(self, item):
        data = {
            "downloader": realize(self.downloader, item),
            "api_version": "2"
        }
        if self.version:
            data["version"] = realize(self.version, item)
        return data

    def process_body(self, body, item):
        data = json.loads(body)
        if "item_name" in data:
            for (k, v) in data.items():
                item[k] = v
            item.log_output(
                "Received item '%s' from tracker\n" % item["item_name"])
            self.complete_item(item)
        else:
            item.log_output("Tracker responded with empty response.\n")
            self.schedule_retry(item)


class SendDoneToTracker(TrackerRequest):
    '''Inform the Tracker the work unit has been completed.'''
    def __init__(self, tracker_url, stats):
        TrackerRequest.__init__(self, "SendDoneToTracker", tracker_url, "done")
        self.stats = stats

    def data(self, item):
        return realize(self.stats, item)

    def process_body(self, body, item):
        if body.strip() == "OK":
            item.log_output(
                "Tracker confirmed item '%s'.\n" % item["item_name"])
            self.complete_item(item)
        else:
            item.log_output(
                "Tracker responded with unexpected '%s'.\n" % body.strip())
            self.schedule_retry(item)


class PrepareStatsForTracker(SimpleTask):
    '''Apply statistical values on the item.'''
    def __init__(self, defaults=None, file_groups=None, id_function=None):
        SimpleTask.__init__(self, "PrepareStatsForTracker")
        self.defaults = defaults or {}
        self.file_groups = file_groups or {}
        self.id_function = id_function

    def process(self, item):
        total_bytes = {}
        for (group, files) in self.file_groups.items():
            total_bytes[group] = sum(
                [os.path.getsize(realize(f, item)) for f in files]
            )

        stats = {}
        stats.update(self.defaults)
        stats["item"] = item["item_name"]
        stats["bytes"] = total_bytes

        if self.id_function:
            stats["id"] = self.id_function(item)

        item["stats"] = realize(stats, item)


class UploadWithTracker(TrackerRequest):
    '''Upload work unit results.

    One of the inner task is used depending on the Tracker's response
    to where to upload:

    * :class:`RsyncUpload`
    * :class:`CurlUpload`
    '''
    def __init__(self, tracker_url, downloader, files, version=None,
                 rsync_target_source_path="./", rsync_bwlimit="0",
                 rsync_extra_args=[], curl_connect_timeout="60",
                 curl_speed_limit="1", curl_speed_time="900"):
        TrackerRequest.__init__(self, "Upload", tracker_url, "upload")

        self.downloader = downloader
        self.version = version

        self.files = files
        self.rsync_target_source_path = rsync_target_source_path
        self.rsync_bwlimit = rsync_bwlimit
        self.rsync_extra_args = rsync_extra_args
        self.curl_connect_timeout = curl_connect_timeout
        self.curl_speed_limit = curl_speed_limit
        self.curl_speed_time = curl_speed_time

    def data(self, item):
        data = {"downloader": realize(self.downloader, item),
                "item_name": item["item_name"]}
        if self.version:
            data["version"] = realize(self.version, item)
        return data

    def process_body(self, body, item):
        data = json.loads(body)
        if "upload_target" in data:
            inner_task = None

            if re.match(r"^rsync://.+/$", data["upload_target"]):
                item.log_output(
                    "Uploading with Rsync to %s" % data["upload_target"])
                inner_task = RsyncUpload(
                    data["upload_target"], self.files,
                    target_source_path=self.rsync_target_source_path,
                    bwlimit=self.rsync_bwlimit,
                    extra_args=self.rsync_extra_args,
                    max_tries=1)

            elif re.match(r"^https?://.+/$", data["upload_target"]):
                item.log_output(
                    "Uploading with Curl to %s" % data["upload_target"])

                if len(self.files) != 1:
                    item.log_output("Curl expects to upload a single file.")
                    item.log_output("Contact a tracker admin!")
                    self.schedule_retry(item)
                    return

                inner_task = CurlUpload(
                    data["upload_target"], self.files[0],
                    self.curl_connect_timeout, self.curl_speed_limit,
                    self.curl_speed_time, max_tries=1)

            else:
                item.log_output("Received invalid upload URI {0}."
                                .format(data["upload_target"]))
                item.log_output("Contact a tracker admin!")
                self.schedule_retry(item)
                return

            inner_task.on_complete_item += self._inner_task_complete_item
            inner_task.on_fail_item += self._inner_task_fail_item
            inner_task.enqueue(item)

        else:
            item.log_output("Tracker did not provide an upload target.")
            self.schedule_retry(item)

    def _inner_task_complete_item(self, task, item):
        self.complete_item(item)

    def _inner_task_fail_item(self, task, item):
        self.schedule_retry(item)



import unittest
import subprocess
import sys


class RunPipelineTest(unittest.TestCase):
    def test_example_pipeline(self):
        if sys.version_info[0] == 3:
            python_exe = 'python3'
            pipeline_exe = './run-pipeline3'
        else:
            python_exe = 'python'
            pipeline_exe = './run-pipeline'

        subprocess.check_call([
            python_exe,
            pipeline_exe,
            './examples/example-pipeline.py',
            'testuser',
            '--max-items', '1',
            '--disable-web-server'
        ])




import time
import os
import os.path
import shutil
import json

from seesaw.project import *
from seesaw.item import *
from seesaw.config import *
from seesaw.task import *
from seesaw.pipeline import *
from seesaw.externalprocess import *
from seesaw.tracker import *

DATA_DIR = "data"
USER_AGENT = "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27"
VERSION = "20120603.01"

class PrepareDirectories(SimpleTask):
  def __init__(self):
    SimpleTask.__init__(self, "PrepareDirectories")

  def process(self, item):
    item_id8 = "%08d" % int(item["item_name"])
    prefix_dir = "/".join(( DATA_DIR, item_id8[7:8], item_id8[6:8], item_id8[5:8] ))
    dirname = "/".join(( prefix_dir, item_id8 ))

    if os.path.isdir(dirname):
      shutil.rmtree(dirname)

    os.makedirs(dirname + "/files")

    item["item_dir"] = dirname
    item["prefix_dir"] = prefix_dir
    item["warc_file_base"] = "picplz-%s-%s" % (item_id8, time.strftime("%Y%m%d-%H%M%S"))

class MoveFiles(SimpleTask):
  def __init__(self):
    SimpleTask.__init__(self, "MoveFiles")

  def process(self, item):
    os.rename("%(item_dir)s/%(warc_file_base)s.warc.gz" % item,
              "%(prefix_dir)s/%(warc_file_base)s.warc.gz" % item)
    os.rename("%(item_dir)s/%(warc_file_base)s.json" % item,
              "%(prefix_dir)s/%(warc_file_base)s.json" % item)

def calculate_item_id(item):
  with open("%s/%s.json" % (item["item_dir"], item["warc_file_base"])) as fp:
    return json.load(fp)

project = Project(
  title = "Picplz",
  project_html = """
      <img class="project-logo" alt="Picplz" src="https://s3.amazonaws.com/data.tumblr.com/tumblr_l3vf57DJ1e1qaewyu.png" height="50px" />
      <h2>Picplz <span class="links"><a href="http://picplz.com/">Picplz website</a> &middot; <a href="http://picplz.heroku.com/">Leaderboard</a></span></h2>
      <p>picplz is a photo sharing app that makes it easy for you to share your mobile pictures on the Web with just a few clicks.</p>
  """,
  utc_deadline = datetime.datetime(2013,1,1, 12,0,0)
)

pipeline = Pipeline(
  GetItemFromTracker("http://picplz-3.herokuapp.com", downloader),
  PrepareDirectories(),
  LimitConcurrent(4,
    WgetDownload([ "./wget-warc-lua",
      "-U", USER_AGENT,
      "-nv",
      "-o", ItemInterpolation("%(item_dir)s/wget.log"),
      "--lua-script", "picplz-user.lua",
      "--directory-prefix", ItemInterpolation("%(item_dir)s/files"),
      "--force-directories",
      "-e", "robots=off",
      "--page-requisites", "--span-hosts",
      "--warc-file", ItemInterpolation("%(item_dir)s/%(warc_file_base)s"),
      "--warc-header", "operator: Archive Team",
      "--warc-header", "picplz-dld-script-version: " + VERSION,
      "--warc-header", ItemInterpolation("picplz-user-id: %(item_name)s"),
      ItemInterpolation("http://api.picplz.com/api/v2/user.json?id=%(item_name)s&include_detail=1&include_pics=1&pic_page_size=100")
    ],
    max_tries = 2,
    accept_on_exit_code = [ 0, 8 ],
#   retry_on_exit_code = [ 1 ],
    env = { "picplz_lua_json": ItemInterpolation("%(item_dir)s/%(warc_file_base)s.json") })
  ),
  PrepareStatsForTracker(
    defaults = { "downloader": downloader, "version": VERSION },
    file_groups = {
      "user": [ ItemInterpolation("%(item_dir)s/%(warc_file_base)s.warc.gz") ]
    },
    id_function = calculate_item_id
  ),
  MoveFiles(),
  LimitConcurrent(1,
    RsyncUpload(
      target = "fos.textfiles.com::picplz/%s/" % downloader,
      target_source_path = "data/",
      files = [
        ItemInterpolation("%(prefix_dir)s/%(warc_file_base)s.warc.gz"),
        ItemInterpolation("%(prefix_dir)s/%(warc_file_base)s.json")
      ],
      bwlimit=ConfigValue(name="Rsync bwlimit", default="0")
    ),
  ),
  SendDoneToTracker(
    tracker_url = "http://picplz-3.herokuapp.com",
    stats = ItemValue("stats")
  )
)




from __future__ import print_function

import argparse
from argparse import ArgumentParser

import seesaw
seesaw.runner_type = "Warrior"

from seesaw.warrior import Warrior
from seesaw.web import start_warrior_server


def main():
    parser = ArgumentParser(description="Run the warrior web interface")
    parser.add_argument("--projects-dir", dest="projects_dir",
                        metavar="DIRECTORY", type=str,
                        help="the warrior projects directory", required=True)
    parser.add_argument("--data-dir", dest="data_dir", metavar="DIRECTORY",
                        type=str,
                        help="the data directory", required=True)
    parser.add_argument("--warrior-hq", dest="warrior_hq_url", metavar="URL",
                        type=str,
                        help="the url to the Warrior HQ", required=True)
    parser.add_argument("--address", dest="address",
                        help="the IP address of the web interface "
                             "(default: 0.0.0.0)",
                        metavar="HOST", type=str, default="0.0.0.0")
    parser.add_argument("--port", dest="port_number",
                        help="the port number for the web interface "
                             "(default: 8001)",
                        metavar="PORT", type=int, default=8001)
    parser.add_argument("--http-username", dest="http_username",
                        help="username for the web interface",
                        metavar="USERNAME", type=str)
    parser.add_argument("--http-password", dest="http_password",
                        help="username for the web interface (default: admin)",
                        metavar="PASSWORD", type=str
                        )  # default is set in start_warrior_server
    parser.add_argument("--keep-data", dest="keep_data",
                        help="do not remove data of finished items",
                        action="store_true")
    parser.add_argument("--real-shutdown", dest="real_shutdown",
                        help="the shutdown button in the web interface uses "
                             "sudo shutdown",
                        action="store_true")
    # extra option to report the warrior VM version to the tracker
    # ask before using
    parser.add_argument("--warrior-build", dest="warrior_build",
                        help=argparse.SUPPRESS, type=str)
    args = parser.parse_args()

    if args.warrior_build:
        seesaw.warrior_build = args.warrior_build

    print("*" * 74)
    print("*%-072s*" % " ")
    print("*%-072s*" % ("   ArchiveTeam Seesaw kit - %s" % seesaw.__version__))
    print("*%-072s*" % " ")
    print("*" * 74)
    print()
    print("Starting the web interface on %s:%d..." %
          (args.address, args.port_number))
    print()
    print("-" * 74)
    print()

    warrior = Warrior(
        args.projects_dir,
        args.data_dir,
        args.warrior_hq_url,
        real_shutdown=args.real_shutdown,
        keep_data=args.keep_data)

    start_warrior_server(warrior,
                         bind_address=args.address,
                         port_number=args.port_number,
                         http_username=args.http_username,
                         http_password=args.http_password)

    warrior.start()


if __name__ == "__main__":
    main()



'''Pipeline execution.'''
import datetime
import functools
import os.path
import sys

import seesaw.util
from seesaw.config import realize
from seesaw.event import Event
from seesaw.item import Item

from tornado import ioloop


class Runner(object):
    '''Executes and manages the lifetime of :class:`Pipeline` instances.'''
    def __init__(self, stop_file=None, concurrent_items=1, max_items=None,
                 keep_data=False):
        self.pipeline = None
        self.concurrent_items = concurrent_items
        self.max_items = max_items
        self.keep_data = keep_data

        self.item_count = 0
        self.active_items = set()
        self.stop_flag = False
        self.stop_file = stop_file
        self.initial_stop_file_mtime = self.stop_file_mtime()

        self.on_status = Event()
        self.on_create_item = Event()
        self.on_pipeline_start_item = Event()
        self.on_pipeline_finish_item = Event()
        self.on_finish = Event()

        if stop_file:
            ioloop.PeriodicCallback(self.check_stop_file, 5000).start()

    def set_current_pipeline(self, pipeline):
        old_pipeline = self.pipeline

        if pipeline:
            pipeline.on_start_item += self._item_starting
            pipeline.on_finish_item += self._item_finished

        self.pipeline = pipeline

        if old_pipeline:
            # stop any cancellable items in the previous pipeline
            old_pipeline.cancel_items()

    def is_active(self):
        return len(self.active_items) > 0

    def start(self):
        self.add_items()

    def stop_gracefully(self):
        print("Stopping when current tasks are completed...")
        self.stop_flag = True
        self.pipeline.cancel_items()
        self.initial_stop_file_mtime = self.stop_file_mtime()
        self.on_status(self, "stopping")

    def keep_running(self):
        print("Keep running...")
        self.stop_flag = False
        self.initial_stop_file_mtime = self.stop_file_mtime()
        self.on_status(self, "running")

    def check_stop_file(self):
        if self.stop_file_changed():
            self.stop_gracefully()

    def should_stop(self):
        return self.stop_flag or self.stop_file_changed()

    def stop_file_changed(self):
        current_stop_file_mtime = self.stop_file_mtime()
        if current_stop_file_mtime:
            return self.initial_stop_file_mtime is None \
                or self.initial_stop_file_mtime < current_stop_file_mtime
        else:
            return False

    def stop_file_mtime(self):
        if self.stop_file and os.path.exists(self.stop_file):
            return os.path.getmtime(self.stop_file)
        else:
            return None

    def add_items(self):
        if self.pipeline:
            items_required = int(realize(self.concurrent_items))
            while len(self.active_items) < items_required:
                if self.max_items and self.max_items <= self.item_count:
                    return

                self.item_count += 1
                item_id = "{0}-{1}".format(
                    seesaw.util.unique_id_str(), self.item_count)
                item = Item(
                    pipeline=self.pipeline,
                    item_id=item_id,
                    item_number=self.item_count,
                    keep_data=self.keep_data
                )
                self.on_create_item(self, item)
                self.active_items.add(item)
                self.pipeline.enqueue(item)

    def _item_starting(self, pipeline, item):
        self.on_pipeline_start_item(self, pipeline, item)

    def _item_finished(self, pipeline, item):
        if item.failed:
            item.log_output("Waiting 10 seconds...")
            ioloop.IOLoop.instance().add_timeout(
                datetime.timedelta(seconds=10),
                functools.partial(
                    self._item_finished_without_delay, pipeline, item)
            )
        else:
            self._item_finished_without_delay(pipeline, item)

    def _item_finished_without_delay(self, pipeline, item):
        self.on_pipeline_finish_item(self, pipeline, item)
        self.active_items.remove(item)

        def add_more_items():
            if not self.should_stop():
                self.add_items()

            if len(self.active_items) == 0:
                self.on_finish(self)

        ioloop.IOLoop.instance().add_timeout(
            datetime.timedelta(),
            add_more_items
        )


class SimpleRunner(Runner):
    '''Executes a single class:`Pipeline` instance.'''
    def __init__(self, pipeline, stop_file=None, concurrent_items=1,
                 max_items=None, keep_data=False):
        Runner.__init__(
            self, stop_file=stop_file,
            concurrent_items=concurrent_items, max_items=max_items,
            keep_data=keep_data)

        self.set_current_pipeline(pipeline)
        self.on_create_item += self._handle_create_item
        self.on_finish += self._stop_ioloop

    def start(self):
        Runner.start(self)
        ioloop.IOLoop.instance().start()
        self.pipeline.on_cleanup()

    def _stop_ioloop(self, dummy):
        ioloop.IOLoop.instance().stop()

    def forced_stop(self):
        print("Stopping immediately...")
        # TODO perhaps the subprocesses should be killed
        ioloop.IOLoop.instance().stop()

    def _handle_create_item(self, dummy, item):
        item.on_output += self._handle_item_output

    def _handle_item_output(self, item, data):
        try:
            sys.stdout.write(data)
        except UnicodeError:
            sys.stdout.write(data.encode('ascii', 'replace').decode('ascii'))
        sys.stdout.flush()



import unittest

import seesaw
import seesaw.six
from seesaw.util import find_executable, unique_id_str


class UtilTest(unittest.TestCase):
    def test_find_executable(self):
        if seesaw.six.PY3:
            exes = ['./run-pipeline3', '../run-pipeline3']
        else:
            exes = ['./run-pipeline', '../run-pipeline']

        self.assertTrue(find_executable(
            'pipeline runner',
            seesaw.__version__,
            exes,
            version_arg='--version')
        )

    def test_find_executable_bad_version(self):
        if seesaw.six.PY3:
            exes = ['./run-pipeline3', '../run-pipeline3']
        else:
            exes = ['./run-pipeline', '../run-pipeline']

        self.assertFalse(find_executable(
            'pipeline runner',
            '123-notrealversion',
            exes,
            version_arg='--version')
        )

    def test_unique_id_str(self):
        # check for no crash
        self.assertTrue(unique_id_str())



'''The warrior server.

The warrior phones home to Warrior HQ
(https://github.com/ArchiveTeam/warrior-hq).
'''
import datetime
from distutils.version import StrictVersion
import json
import os
import os.path
import re
import shutil
import subprocess
import sys
import time

from tornado import gen
from tornado import ioloop
from tornado.httpclient import AsyncHTTPClient

import seesaw
from seesaw.config import NumberConfigValue, StringConfigValue, ConfigValue
from seesaw.config import realize
from seesaw.event import Event
from seesaw.externalprocess import AsyncPopen
from seesaw.runner import Runner
import seesaw.six


try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict


if seesaw.six.PY2:
    bigint = long  # @UndefinedVariable  pylint: disable=undefined-variable
else:
    bigint = int


class ConfigManager(object):
    '''Manages the configuration.'''
    def __init__(self, config_file):
        self.config_file = config_file
        self.config_memory = {}
        self.config_values = OrderedDict()

        self.load()

    def add(self, config_value):
        self.config_values[config_value.name] = config_value
        if config_value.name in self.config_memory:
            remembered_value = self.config_memory[config_value.name]
            if config_value.check_value(remembered_value) is None:
                config_value.set_value(remembered_value)
        self.save()

    def remove(self, name):
        if name in self.config_values:
            del self.config_values[name]
        self.save()

    def all_valid(self):
        return all([c.is_valid() for c in self])

    def set_value(self, name, value):
        if name in self.config_values:
            if self.config_values[name].set_value(value):
                self.config_memory[name] = value
                self.save()
                return True
        return False

    def load(self):
        try:
            with open(self.config_file) as f:
                self.config_memory = json.load(f)
        except Exception:
            self.config_memory = {}

    def save(self):
        with open(self.config_file, "w") as f:
            json.dump(self.config_memory, f)

    def __iter__(self):
        return iter(self.config_values.values())

    def editable_values(self):
        return [v for v in self if v.editable]


class BandwidthMonitor(object):
    '''Extracts the bandwidth usage from the system stats.'''
    devre = re.compile(r"^\s*([a-z0-9]+):(.+)$")

    def __init__(self, device):
        self.device = device
        self.prev_time = None
        self.prev_stats = None
        self.bandwidth = None
        self._prev_received = 0
        self._prev_sent = 0
        self._overflow_received = 0
        self._overflow_sent = 0

        self.update()

    def current_stats(self):
        if self.prev_stats and self.bandwidth:
            return {
                "received": self.prev_stats[0],
                "sent": self.prev_stats[1],
                "receiving": self.bandwidth[0],
                "sending": self.bandwidth[1]
            }
        return None

    def update(self):
        cur_time = time.time()
        cur_stats = self._get_stats()
        if self.prev_stats is not None and cur_stats is not None:
            time_delta = cur_time - self.prev_time
            if time_delta:
                self.bandwidth = [
                    (cur_stats[0] - self.prev_stats[0]) / time_delta,
                    (cur_stats[1] - self.prev_stats[1]) / time_delta,
                ]
        self.prev_time = cur_time
        self.prev_stats = cur_stats
        return self.bandwidth

    def _get_stats(self):
        with open("/proc/net/dev") as f:
            lines = f.readlines()
        for line in lines:
            m = self.devre.match(line)
            if m and m.group(1) == self.device:
                fields = m.group(2).split()
                received = bigint(fields[0])
                sent = bigint(fields[8])
                if self._prev_received > received:
                    self._overflow_received += 2 ** 32
                self._prev_received = received
                if self._prev_sent > sent:
                    self._overflow_sent += 2 ** 32
                self._prev_sent = sent
                return [received + self._overflow_received,
                        sent + self._overflow_sent ]
        return None


class Warrior(object):
    '''The warrior god object.'''
    def __init__(self, projects_dir, data_dir, warrior_hq_url,
                 real_shutdown=False, keep_data=False):
        if not os.access(projects_dir, os.W_OK):
            raise Exception(
                "Couldn't write to projects directory: %s" % projects_dir)
        if not os.access(data_dir, os.W_OK):
            raise Exception("Couldn't write to data directory: %s" % data_dir)

        self.projects_dir = projects_dir
        self.data_dir = data_dir
        self.warrior_hq_url = warrior_hq_url
        self.real_shutdown = real_shutdown
        self.keep_data = keep_data

        # disable the password prompts
        self.gitenv = dict(
            list(os.environ.items()) +
            list({
                'GIT_ASKPASS': 'echo',
                'SSH_ASKPASS': 'echo'
            }.items())
        )

        self.warrior_id = StringConfigValue(
            name="warrior_id",
            title="Warrior ID",
            description="The unique number of your warrior instance.",
            editable=False
        )
        self.selected_project_config_value = StringConfigValue(
            name="selected_project",
            title="Selected project",
            description="The project (to be continued when the warrior "
                        "restarts).",
            default="none",
            editable=False
        )
        self.downloader = StringConfigValue(
            name="downloader",
            title="Your nickname",
            description="We use your nickname to show your results on our "
                        "tracker. Letters and numbers only.",
            regex="^[-_a-zA-Z0-9]{3,30}$",
            advanced=False
        )
        self.concurrent_items = NumberConfigValue(
            name="concurrent_items",
            title="Concurrent items",
            description="How many items should the warrior download at a "
                        "time? (Max: 6)",
            min=1,
            max=6,
            default=2
        )
        self.http_username = StringConfigValue(
            name="http_username",
            title="HTTP username",
            description="Enter a username to protect the web interface, "
                        "or leave empty.",
            default=""
        )
        self.http_password = StringConfigValue(
            name="http_password",
            title="HTTP password",
            description="Enter a password to protect the web interface, "
                        "or leave empty.",
            default=""
        )

        self.config_manager = ConfigManager(os.path.join(projects_dir,
                                                         "config.json"))
        self.config_manager.add(self.warrior_id)
        self.config_manager.add(self.selected_project_config_value)
        self.config_manager.add(self.downloader)
        self.config_manager.add(self.concurrent_items)
        self.config_manager.add(self.http_username)
        self.config_manager.add(self.http_password)

        self.bandwidth_monitor = BandwidthMonitor("eth0")
        self.bandwidth_monitor.update()

        self.runner = Runner(concurrent_items=self.concurrent_items,
                             keep_data=self.keep_data)
        self.runner.on_finish += self.handle_runner_finish

        self.current_project_name = None
        self.current_project = None

        self.selected_project = None

        self.projects = {}
        self.installed_projects = set()
        self.failed_projects = set()

        self.on_projects_loaded = Event()
        self.on_project_installing = Event()
        self.on_project_installed = Event()
        self.on_project_installation_failed = Event()
        self.on_project_refresh = Event()
        self.on_project_selected = Event()
        self.on_status = Event()

        self.http_client = AsyncHTTPClient()

        self.installing = False
        self.shut_down_flag = False
        self.reboot_flag = False

        self.hq_updater = ioloop.PeriodicCallback(self.update_warrior_hq,
                                                  10 * 60 * 1000)
        self.project_updater = ioloop.PeriodicCallback(self.update_project,
                                                       60 * 60 * 1000)
        self.forced_reboot_timeout = None

        self.lat_lng = None
        self.find_lat_lng()

        self.install_output = None

    def find_lat_lng(self):
        # response = self.http_client.fetch("http://www.maxmind.com/app/mylocation", self.handle_lat_lng, user_agent="")
        pass

    def handle_lat_lng(self, response):
        m = re.search(r"geoip-demo-results-tbodyLatitude/Longitude</td>"
                      r"\s*<td[^>]*>\s*([-/.0-9]+)\s*</td>",
                      response.body)
        if m:
            self.lat_lng = m.group(1)

    def bandwidth_stats(self):
        self.bandwidth_monitor.update()
        return self.bandwidth_monitor.current_stats()

    @gen.engine
    def update_warrior_hq(self):
        if realize(self.warrior_id) is None:
            headers = {"Content-Type": "application/json"}
            user_agent = "ArchiveTeam Warrior/%s" % seesaw.__version__
            body = json.dumps(
                {"warrior": {"version": seesaw.__version__}}
            )
            response = yield gen.Task(self.http_client.fetch,
                                      os.path.join(self.warrior_hq_url,
                                                   "api/register.json"),
                                      method="POST",
                                      headers=headers,
                                      user_agent=user_agent,
                                      body=body)
            if response.code == 200:
                data = json.loads(response.body.decode('utf-8'))
                print("Received Warrior ID '%s'." % data["warrior_id"])
                self.config_manager.set_value("warrior_id", data["warrior_id"])
            else:
                print("HTTP error %s" % (response.code))
                return
        else:
            print("Warrior ID '%s'." % realize(self.warrior_id))

        headers = {"Content-Type": "application/json"}
        user_agent = "ArchiveTeam Warrior/%s %s" % (seesaw.__version__,
                                                    seesaw.runner_type)
        body = json.dumps({
            "warrior": {
                "warrior_id": realize(self.warrior_id),
                "lat_lng": self.lat_lng,
                "downloader": realize(self.downloader),
                "selected_project": realize(self.selected_project_config_value)
            }})

        response = yield gen.Task(self.http_client.fetch,
                                  os.path.join(self.warrior_hq_url,
                                               "api/update.json"),
                                  method="POST",
                                  headers=headers,
                                  user_agent=user_agent,
                                  body=body)
        if response.code == 200:
            data = json.loads(response.body.decode('utf-8'))

            if StrictVersion(seesaw.__version__) < \
                    StrictVersion(data["warrior"]["seesaw_version"]):
                # time for an update
                print("Reboot for Seesaw update.")
                self.reboot_gracefully()

                # schedule a forced reboot after two days
                self.schedule_forced_reboot()
                return

            projects_list = data["projects"]
            self.projects = OrderedDict(
                [(project["name"], project) for project in projects_list])
            for project_data in self.projects.values():
                if "deadline" in project_data:
                    project_data["deadline_int"] = time.mktime(
                        time.strptime(project_data["deadline"],
                                      "%Y-%m-%dT%H:%M:%SZ"))

            previous_project_choice = realize(
                self.selected_project_config_value)

            if self.selected_project and \
                    self.selected_project not in self.projects:
                self.select_project(None)
            elif previous_project_choice in self.projects:
                # select previous project
                self.select_project(previous_project_choice)
            elif previous_project_choice == "auto":
                # ArchiveTeam's choice
                if "auto_project" in data:
                    self.select_project(data["auto_project"])
                else:
                    self.select_project(None)

            self.on_projects_loaded(self, self.projects)

        else:
            print("HTTP error %s" % (response.code))

    @gen.engine
    def install_project(self, project_name, callback=None):
        self.installed_projects.discard(project_name)

        if project_name in self.projects and not self.installing:
            self.installing = project_name
            self.install_output = []

            project = self.projects[project_name]
            project_path = os.path.join(self.projects_dir, project_name)

            self.on_project_installing(self, project)

            if project_name in self.failed_projects:
                if os.path.exists(project_path):
                    shutil.rmtree(project_path)
                self.failed_projects.discard(project_name)

            if os.path.exists(project_path):
                subprocess.Popen(
                    args=["git", "config", "remote.origin.url",
                          project["repository"]],
                    cwd=project_path
                ).communicate()

                p = AsyncPopen(
                    args=["git", "pull"],
                    cwd=project_path,
                    env=self.gitenv
                )
            else:
                p = AsyncPopen(
                    args=["git", "clone", project["repository"], project_path],
                    env=self.gitenv
                )
            p.on_output += self.collect_install_output
            p.on_end += yield gen.Callback("gitend")
            p.run()
            result = yield gen.Wait("gitend")

            if result != 0:
                self.install_output.append("\ngit returned %d\n" % result)
                self.on_project_installation_failed(
                    self, project, "".join(self.install_output))
                self.installing = None
                self.failed_projects.add(project_name)
                if callback:
                    callback(False)
                return

            project_install_file = os.path.join(project_path,
                                                "warrior-install.sh")

            if os.path.exists(project_install_file):
                p = AsyncPopen(
                    args=[project_install_file],
                    cwd=project_path
                )
                p.on_output += self.collect_install_output
                p.on_end += yield gen.Callback("installend")
                p.run()
                result = yield gen.Wait("installend")

                if result != 0:
                    self.install_output.append(
                        "\nCustom installer returned %d\n" % result)
                    self.on_project_installation_failed(
                        self, project, "".join(self.install_output))
                    self.installing = None
                    self.failed_projects.add(project_name)
                    if callback:
                        callback(False)
                    return

            data_dir = os.path.join(self.data_dir, "data")
            if os.path.exists(data_dir):
                shutil.rmtree(data_dir)
            os.makedirs(data_dir)

            project_data_dir = os.path.join(project_path, "data")
            if os.path.islink(project_data_dir):
                os.remove(project_data_dir)
            elif os.path.isdir(project_data_dir):
                shutil.rmtree(project_data_dir)
            os.symlink(data_dir, project_data_dir)

            self.installed_projects.add(project_name)
            self.on_project_installed(self, project,
                                      "".join(self.install_output))

            self.installing = None
            if callback:
                callback(True)

    @gen.engine
    def update_project(self):
        if self.selected_project and (yield gen.Task(
                self.check_project_has_update, self.selected_project)):
            # restart project
            self.start_selected_project()

    @gen.engine
    def check_project_has_update(self, project_name, callback):
        if project_name in self.projects:
            project = self.projects[project_name]
            project_path = os.path.join(self.projects_dir, project_name)

            self.install_output = []

            if not os.path.exists(project_path):
                callback(True)
                return

            subprocess.Popen(
                args=["git", "config", "remote.origin.url",
                      project["repository"]],
                cwd=project_path
            ).communicate()

            p = AsyncPopen(
                args=["git", "fetch"],
                cwd=project_path,
                env=self.gitenv
            )
            p.on_output += self.collect_install_output
            p.on_end += yield gen.Callback("gitend")
            p.run()
            result = yield gen.Wait("gitend")

            if result != 0:
                callback(True)
                return

            output = subprocess.Popen(
                args=["git", "rev-list", "HEAD..FETCH_HEAD"],
                cwd=project_path,
                stdout=subprocess.PIPE
            ).communicate()[0]
            if output.strip() != "":
                callback(True)
            else:
                callback(False)

    def collect_install_output(self, data):
        sys.stdout.write(data)
        data = re.sub("[\x00-\x08\x0b\x0c]", "", data)
        self.install_output.append(data)

    @gen.engine
    def select_project(self, project_name):
        if project_name == "auto":
            self.update_warrior_hq()
            return

        if project_name not in self.projects:
            project_name = None

        if project_name != self.selected_project:
            # restart
            self.selected_project = project_name
            self.on_project_selected(self, project_name)
            self.start_selected_project()

    def clone_project(self, project_name, project_path):
        version_string = subprocess.Popen(
            args=["git", "log", "-1", "--pretty=%h"],
            cwd=project_path,
            stdout=subprocess.PIPE
        ).communicate()[0].strip().decode('ascii')

        project_versioned_path = os.path.join(
            self.data_dir, "projects",
            "%s-%s" % (project_name, version_string))
        if not os.path.exists(project_versioned_path):
            if not os.path.exists(os.path.join(self.data_dir, "projects")):
                os.makedirs(os.path.join(self.data_dir, "projects"))

            subprocess.Popen(
                args=["git", "clone", project_path, project_versioned_path],
                env=self.gitenv
            ).communicate()

        return project_versioned_path

    def load_pipeline(self, pipeline_path, context):
        dirname, basename = os.path.split(pipeline_path)
        if dirname == "":
            dirname = "."

        with open(pipeline_path) as f:
            pipeline_str = f.read()

        ConfigValue.start_collecting()

        local_context = context
        global_context = context
        curdir = os.getcwd()
        try:
            os.chdir(dirname)
            exec(pipeline_str, local_context, global_context)
        finally:
            os.chdir(curdir)

        config_values = ConfigValue.stop_collecting()

        project = local_context["project"]
        pipeline = local_context["pipeline"]
        pipeline.project = project
        return (project, pipeline, config_values)

    @gen.engine
    def start_selected_project(self):
        project_name = self.selected_project

        if project_name in self.projects:
            # install or update project if necessary
            if project_name not in self.installed_projects or \
                    (yield gen.Task(self.check_project_has_update,
                                    project_name)):
                result = yield gen.Task(self.install_project, project_name)
                if not result:
                    return

            # remove the configuration variables from the previous project
            if self.current_project:
                for config_value in self.current_project.config_values:
                    self.config_manager.remove(config_value.name)

            # the path with the project code
            # (this is the most recent code from the repository)
            project_path = os.path.join(self.projects_dir, project_name)

            # clone the project code to a versioned directory
            # where the pipeline is actually run
            project_versioned_path = self.clone_project(project_name,
                                                        project_path)

            # load the pipeline from the versioned directory
            pipeline_path = os.path.join(project_versioned_path, "pipeline.py")
            (project, pipeline, config_values) = self.load_pipeline(
                pipeline_path, {"downloader": self.downloader})

            # add the configuration values to the config manager
            for config_value in config_values:
                self.config_manager.add(config_value)
            project.config_values = config_values

            # start the pipeline
            if not self.shut_down_flag and not self.reboot_flag:
                self.runner.set_current_pipeline(pipeline)

            self.current_project_name = project_name
            self.current_project = project

            self.on_project_refresh(self, self.current_project, self.runner)
            self.fire_status()

            if not self.shut_down_flag and not self.reboot_flag:
                self.runner.start()

        else:
            # project_name not in self.projects,
            # stop the current project (if there is one)
            self.runner.set_current_pipeline(None)
            self.fire_status()

    def handle_runner_finish(self, runner):
        if self.current_project:
            for config_value in self.current_project.config_values:
                self.config_manager.remove(config_value.name)

        self.current_project_name = None
        self.current_project = None

        self.on_project_refresh(self, self.current_project, self.runner)
        self.fire_status()

        if self.shut_down_flag or self.reboot_flag:
            ioloop.IOLoop.instance().stop()

            if self.real_shutdown:
                if self.shut_down_flag:
                    os.system("sudo shutdown -h now")
                elif self.reboot_flag:
                    os.system("sudo shutdown -r now")

    def start(self):
        if self.real_shutdown:
            # schedule a reboot
            ioloop.IOLoop.instance().add_timeout(datetime.timedelta(days=7),
                                                 self.max_age_reached)

        self.hq_updater.start()
        self.project_updater.start()
        self.update_warrior_hq()
        ioloop.IOLoop.instance().start()

    def max_age_reached(self):
        if self.real_shutdown:
            # time for an sanity reboot
            print("Running for more than 7 days. Time to schedule a reboot.")
            self.reboot_gracefully()

            # schedule a forced reboot after two days
            self.schedule_forced_reboot()

    def reboot_gracefully(self):
        self.shut_down_flag = False
        self.reboot_flag = True
        self.fire_status()
        if self.runner.is_active():
            self.runner.set_current_pipeline(None)
        else:
            ioloop.IOLoop.instance().stop()
            if self.real_shutdown:
                os.system("sudo shutdown -r now")

    def schedule_forced_reboot(self):
        if self.real_shutdown and not self.forced_reboot_timeout:
            self.forced_reboot_timeout = ioloop.IOLoop.instance().add_timeout(
                datetime.timedelta(days=2), self.forced_reboot)

    def forced_reboot(self):
        print("Stopping immediately...")
        if self.real_shutdown:
            os.system("sudo shutdown -r now")

    def stop_gracefully(self):
        self.shut_down_flag = True
        self.reboot_flag = False
        self.fire_status()
        if self.runner.is_active():
            self.runner.set_current_pipeline(None)
        else:
            ioloop.IOLoop.instance().stop()
            if self.real_shutdown:
                os.system("sudo shutdown -h now")

    def forced_stop(self):
        ioloop.IOLoop.instance().stop()
        if self.real_shutdown:
            os.system("sudo shutdown -h now")

    def keep_running(self):
        self.shut_down_flag = False
        self.reboot_flag = False
        self.start_selected_project()
        self.fire_status()

    class Status(object):
        NO_PROJECT = "NO_PROJECT"
        INVALID_SETTINGS = "INVALID_SETTINGS"
        STOPPING_PROJECT = "STOPPING_PROJECT"
        RESTARTING_PROJECT = "RESTARTING_PROJECT"
        RUNNING_PROJECT = "RUNNING_PROJECT"
        SWITCHING_PROJECT = "SWITCHING_PROJECT"
        STARTING_PROJECT = "STARTING_PROJECT"
        SHUTTING_DOWN = "SHUTTING_DOWN"
        REBOOTING = "REBOOTING"

    def fire_status(self):
        self.on_status(self, self.warrior_status())

    def warrior_status(self):
        if self.shut_down_flag:
            return Warrior.Status.SHUTTING_DOWN
        elif self.reboot_flag:
            return Warrior.Status.REBOOTING
        elif not self.config_manager.all_valid():
            return Warrior.Status.INVALID_SETTINGS
        elif self.selected_project is None and \
                self.current_project_name is None:
            return Warrior.Status.NO_PROJECT
        elif self.selected_project:
            if self.selected_project == self.current_project_name:
                return Warrior.Status.RUNNING_PROJECT
            else:
                return Warrior.Status.STARTING_PROJECT
        else:
            return Warrior.Status.STOPPING_PROJECT



'''Managing steps in a work unit.'''
import contextlib
import os
import traceback

from seesaw.event import Event
from seesaw.item import Item
from seesaw.config import realize


class Task(object):
    '''A step in the download process of an :class:`Item`.
    '''
    def __init__(self, name):
        self.name = name
        self.cwd = os.getcwd()
        self.on_start_item = Event()
        self.on_complete_item = Event()
        self.on_fail_item = Event()
        self.on_finish_item = Event()

    def start_item(self, item):
        item.set_task_status(self, Item.TaskStatus.running)
        self.on_start_item(self, item)

    def fail_item(self, item):
        item.set_task_status(self, Item.TaskStatus.failed)
        self.on_fail_item(self, item)
        self.on_finish_item(self, item)

    def complete_item(self, item):
        item.set_task_status(self, Item.TaskStatus.completed)
        self.on_complete_item(self, item)
        self.on_finish_item(self, item)

    @contextlib.contextmanager
    def task_cwd(self):
        curdir = os.getcwd()
        try:
            os.chdir(self.cwd)
            yield
        finally:
            os.chdir(curdir)

    def fill_ui_task_list(self, task_list):
        task_list.append((self, self.name))

    def __str__(self):
        return self.name


class SimpleTask(Task):
    '''A subclassable :class:`Task` that should do one small thing well.

    Example::

        class MyTask(SimpleTask):
            def process(self, item):
                item['my_message'] = 'hello world!'
    '''
    def __init__(self, name):
        Task.__init__(self, name)

    def enqueue(self, item):
        self.start_item(item)
        item.log_output("Starting %s for %s\n" % (self, item.description()))
        try:
            with self.task_cwd():
                self.process(item)
        except Exception as e:
            item.log_output("Failed %s for %s\n" % (self, item.description()))
            item.log_output("%s\n" % traceback.format_exc())
            item.log_error(self, e)
            self.fail_item(item)
        else:
            item.log_output("Finished %s for %s\n" % (self,
                                                      item.description()))
            self.complete_item(item)

    def process(self, item):
        # TODO: should this raise NotImplemented or be decorated with
        # abc.abstractmethod?
        pass

    def __str__(self):
        return self.name


class LimitConcurrent(Task):
    '''Restricts the number of tasks of the same type that can be run at once.
    '''
    def __init__(self, concurrency, inner_task):
        Task.__init__(self, "LimitConcurrent")
        self.concurrency = concurrency
        self.inner_task = inner_task
        self.inner_task.on_complete_item += self._inner_task_complete_item
        self.inner_task.on_fail_item += self._inner_task_fail_item
        self._queue = []
        self._working = 0

    def enqueue(self, item):
        if self._working < realize(self.concurrency, item):
            self._working += 1
            self.inner_task.enqueue(item)
        else:
            self._queue.append(item)

    def _inner_task_complete_item(self, task, item):
        self._working -= 1
        if len(self._queue) > 0:
            self._working += 1
            self.inner_task.enqueue(self._queue.pop(0))
        self.complete_item(item)

    def _inner_task_fail_item(self, task, item):
        self._working -= 1
        if len(self._queue) > 0:
            self._working += 1
            self.inner_task.enqueue(self._queue.pop(0))
        self.fail_item(item)

    def fill_ui_task_list(self, task_list):
        self.inner_task.fill_ui_task_list(task_list)

    def __str__(self):
        return "LimitConcurrent({0} x {1} )".format(
            self.concurrency, self.inner_task)


class ConditionalTask(Task):
    '''Runs a task optionally.'''
    def __init__(self, condition_function, inner_task):
        Task.__init__(self, "Conditional")
        self.condition_function = condition_function
        self.inner_task = inner_task
        self.inner_task.on_complete_item += self._inner_task_complete_item
        self.inner_task.on_fail_item += self._inner_task_fail_item

    def enqueue(self, item):
        if self.condition_function(item):
            self.inner_task.enqueue(item)
        else:
            item.log_output("Skipping tasks for this item.")
            self.complete_item(item)

    def _inner_task_complete_item(self, task, item):
        self.complete_item(item)

    def _inner_task_fail_item(self, task, item):
        self.fail_item(item)

    def fill_ui_task_list(self, task_list):
        self.inner_task.fill_ui_task_list(task_list)

    def __str__(self):
        return "Conditional(" + str(self.inner_task) + ")"


class SetItemKey(SimpleTask):
    '''Set a value onto a task.'''
    def __init__(self, key, value):
        SimpleTask.__init__(self, "SetItemKey")
        self.key = key
        self.value = value

    def process(self, item):
        item[self.key] = realize(self.value, self)

    def __str__(self):
        return "SetItemKey(" + str(self.key) + ": " + str(self.value) + ")"


class PrintItem(SimpleTask):
    '''Output the name of the :class:`Item`.'''
    def __init__(self):
        SimpleTask.__init__(self, "PrintItem")

    def process(self, item):
        item.log_output("%s\n" % str(item))



'''Project information.'''
import time


class Project(object):
    '''Briefly describes a project metadata.

    This class defines the title of the project, a short description with an
    optional project logo and an optional deadline. The information will be
    shown in the web interface when the project is running.
    '''
    def __init__(self, title=None, project_html=None, utc_deadline=None):
        self.title = title
        self.project_html = project_html
        self.utc_deadline = utc_deadline

    def data_for_json(self):
        return {
            "project_id": id(self),
            "title": self.title,
            "project_html": self.project_html,
            "utc_deadline": (time.mktime(self.utc_deadline.timetuple())
                             if self.utc_deadline else None)
        }



'''Miscellaneous functions.'''
import os
import subprocess
import time
import base64


def test_executable(name, version, path, version_arg="-V"):
    '''Try to run an executable and check its version.'''
    print("Looking for %s in %s" % (name, path))
    try:
        process = subprocess.Popen(
            [path, version_arg],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        stdout_data, stderr_data = process.communicate()
        result = stdout_data + stderr_data
        if not process.returncode == 0:
            print("%s: Returned code %d" % (path, process.returncode))
            return False

        if isinstance(version, str):
            if not version.encode('utf-8') in result:
                print("%s: Incorrect %s version (want %s)." % (path, name,
                                                               version))
                return False
        elif hasattr(version, "search"):
            if not version.search(result):
                print("%s: Incorrect %s version." % (path, name))
                return False
        elif hasattr(version, "__iter__"):
            if not any((v in result) for v in version):
                print("%s: Incorrect %s version (want %s)." % (path, name,
                                                               str(version)))
                return False

        print("Found usable %s in %s" % (name, path))
        return True
    except OSError as e:
        print("%s:" % path, e)
        return False


def find_executable(name, version, paths, version_arg="-V"):
    '''Returns the path of a matching executable.

    .. seealso:: :func:`test_executable`
    '''
    for path in paths:
        if test_executable(name, version, path, version_arg):
            return path
    return None


def unique_id_str():
    '''Returns a unique string suitable for IDs.'''
    rand_str = base64.b16encode(os.urandom(8)).decode('ascii').lower()
    return "{0}{1}".format(int(time.time()), rand_str)



from seesaw.runner import *

DATA_DIR = "data"
USER_AGENT = "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27"
VERSION = "20120603.01"

downloader = "alard"

exec(compile(open("picplz-example-pipeline.py").read(), "picplz-example-pipeline.py", 'exec'))

print(pipeline)

runner = SimpleRunner(pipeline, stop_file="STOP", concurrent_items=1)
runner.start()




# encoding=utf8
from __future__ import unicode_literals

from seesaw.externalprocess import ExternalProcess
from seesaw.pipeline import Pipeline
from seesaw.runner import SimpleRunner
from seesaw.six import StringIO
from seesaw.test_base import BaseTestCase


class ExternalProcessUser(ExternalProcess):
    def __init__(self, *args, **kwargs):
        ExternalProcess.__init__(self, *args, **kwargs)
        self.output_buffer = StringIO()
        self.return_code = None
        self.exit_count = 0
        self.retry_delay = 0.1

    def on_subprocess_stdout(self, pipe, item, data):
        ExternalProcess.on_subprocess_stdout(self, pipe, item, data)
        self.output_buffer.write(data)

    def on_subprocess_end(self, item, returncode):
        ExternalProcess.on_subprocess_end(self, item, returncode)
        self.return_code = returncode
        self.exit_count += 1


class ExternalProcessTest(BaseTestCase):
    def test_proc(self):
        external_process = ExternalProcessUser(
            "Echo", ["python", "-c", "print('hello world!')"], max_tries=4)
        pipeline = Pipeline(external_process)
        pipeline.has_failed = None

        def fail_callback(task, item):
            pipeline.has_failed = True

        pipeline.on_fail_item += fail_callback

        runner = SimpleRunner(pipeline, max_items=1)
        runner.start()

        output = external_process.output_buffer.getvalue()
        self.assertFalse(pipeline.has_failed)
        self.assertTrue('hello world!' in output)
        self.assertIOLoopOK()

    def test_proc_fail(self):
        for max_tries in [1, 2, 20]:
            external_process = ExternalProcessUser(
                "Quitter", ["python", "-c", "import sys;sys.exit(33)"],
                max_tries=max_tries)
            pipeline = Pipeline(external_process)
            pipeline.has_failed = None

            def fail_callback(task, item):
                pipeline.has_failed = True

            pipeline.on_fail_item += fail_callback

            runner = SimpleRunner(pipeline, max_items=1)
            runner.start()

            self.assertTrue(pipeline.has_failed)
            self.assertEqual(33, external_process.return_code)
            self.assertEqual(max_tries, external_process.exit_count)
            self.assertIOLoopOK()

    def test_no_such_file(self):
        external_process = ExternalProcessUser(
            "Fake", ["kitteh and doge.avi.exe"])
        pipeline = Pipeline(external_process)
        pipeline.has_failed = None

        def fail_callback(task, item):
            pipeline.has_failed = True

        pipeline.on_fail_item += fail_callback

        runner = SimpleRunner(pipeline, max_items=1)
        runner.start()
        self.assertTrue(pipeline.has_failed)
        self.assertIOLoopOK()

    def test_proc_stdin_error(self):
        external_process = ExternalProcessUser(
            "Echo", ["python", "-c" "print('hello world!')"], max_tries=4)

        external_process.stdin_data = lambda item: 123456

        pipeline = Pipeline(external_process)
        pipeline.has_failed = None

        def fail_callback(task, item):
            pipeline.has_failed = True

        pipeline.on_fail_item += fail_callback

        runner = SimpleRunner(pipeline, max_items=1)
        runner.start()

        self.assertTrue(pipeline.has_failed)
        self.assertIOLoopOK()
        self.assertEqual(4, external_process.exit_count)

    def test_proc_utf8(self):
        external_process = ExternalProcessUser(
            "Echo", ["python", "-c", "print(u'hello world!áßðfáßðf')"],
        )

        pipeline = Pipeline(external_process)
        pipeline.has_failed = None

        def fail_callback(task, item):
            pipeline.has_failed = True

        pipeline.on_fail_item += fail_callback

        runner = SimpleRunner(pipeline, max_items=1)
        runner.start()

        self.assertFalse(pipeline.has_failed)
        self.assertIOLoopOK()



import unittest

from seesaw.item import Item

class MockPipeline(object):
    pass

class ItemTest(unittest.TestCase):
    def setUp(self):
        self.item = Item(MockPipeline(), 'FakeID', 1, prepare_data_directory=False)

    def test_get_returns_none_for_undefined_keys(self):
        self.assertEquals(None, self.item.get('undefined_key'))

    def test_get_returns_property(self):
        self.item['foo'] = 'bar'

        self.assertEquals('bar', self.item.get('foo'))



'''Configuration value manipulation.'''
import re


def realize(v, item=None):
    '''Makes objects contain concrete values from an item.

    A silly example::

        class AddExpression(object):
            def realize(self, item):
                return = item['x'] + item['y']

        pipeline = Pipeline(ComputeMath(AddExpression()))

    In the example, we want to compute an addition expression. The values
    are defined in the Item.
    '''
    if isinstance(v, dict):
        realized_dict = {}
        for (key, value) in v.items():
            realized_dict[key] = realize(value, item)
        return realized_dict
    elif isinstance(v, list):
        return [realize(vi, item) for vi in v]
    elif hasattr(v, "realize"):
        return v.realize(item)
    else:
        return v


class ConfigValue(object):
    '''Configuration value validator.

    The collection methods are useful for providing user configurable
    settings at run time. For example, when a pipeline file is executed
    by the warrior, the additional config values are presented in the
    warrior configuration panel.
    '''
    collector = None

    @classmethod
    def start_collecting(cls):
        ConfigValue.collector = []

    @classmethod
    def stop_collecting(cls):
        collected = ConfigValue.collector
        ConfigValue.collector = None
        return collected

    def __init__(self, name, title="", description="", default=None,
                 editable=True, advanced=True):
        self.name = name
        self.title = title
        self.description = description
        self.value = self.convert_value(default)
        self.error = None
        self.editable = editable
        self.advanced = advanced

        if ConfigValue.collector is not None:
            ConfigValue.collector.append(self)

    def realize(self, dummy):
        return self.value

    def set_value(self, value):
        self.error = self.check_value(value)
        if self.error is None:
            self.value = self.convert_value(value)
            return True
        else:
            return False

    def check_value(self, value):
        return None

    def convert_value(self, value):
        return value

    def is_valid(self):
        return self.value is not None

    def __str__(self):
        return "<" + self.name + ":" + str(self.value) + ">"


class StringConfigValue(ConfigValue):
    def __init__(self, *args, **kwargs):
        if "regex" in kwargs:
            self.regex = kwargs["regex"]
            del kwargs["regex"]
        else:
            self.regex = None

        ConfigValue.__init__(self, *args, **kwargs)

    def check_value(self, value):
        value = value.strip()
        if self.regex and not re.search(self.regex, value):
            return "Invalid value for %s." % self.title.lower()
        else:
            return None


class NumberConfigValue(ConfigValue):
    def __init__(self, *args, **kwargs):
        if "min" in kwargs:
            self.min = kwargs["min"]
            del kwargs["min"]
        else:
            self.min = None
        if "max" in kwargs:
            self.max = kwargs["max"]
            del kwargs["max"]
        else:
            self.max = None

        ConfigValue.__init__(self, *args, **kwargs)

    def check_value(self, value):
        value = value.strip()
        if not re.match("^[0-9]+$", value):
            return "Invalid number."
        elif self.min and int(value) < self.min:
            return "Number must be %d or greater." % self.min
        elif self.max and int(value) > self.max:
            return "Number must be %d or smaller." % self.max
        else:
            return None

    def convert_value(self, value):
        return int(value)


class ConfigInterpolation(object):
    def __init__(self, s, c):
        self.s = s
        self.c = c

    def realize(self, item):
        return realize(self.s, item) % realize(self.c, item)

    def __str__(self):
        return "<'" + self.s + "'>"



'''Mocks Warrior HQ.'''
import tornado.web


class RegisterHandler(tornado.web.RequestHandler):
    def initialize(self, **kwargs):
        pass

    def post(self):
        self.write({
            'warrior_id': 'test-warrior-id',
        })


class UpdateHandler(tornado.web.RedirectHandler):
    def initialize(self, **kwargs):
        pass

    def post(self):
        self.write({
            "warrior": {
              "seesaw_version": "0.0.15"
           },
           "auto_project": "testproject",
           "projects": [
              {
                 "name": "testproject",
                 "title": "A test project",
                 "description": "Testing a project",
                 "repository":
                    "https://github.com/ArchiveTeam/example-seesaw-project",
                 "logo":
                    "https://raw.github.com/ArchiveTeam/warrior-preseed/master"
                    "/splash/Archive_team-white.png",
                 "marker_html": "hi",
                 "lat_lng": [
                    0.0,
                    0.0
                 ],
                 "leaderboard": "http://example.com/"
              },
           ]
       })


if __name__ == '__main__':
    handlers = [
        (r'/api/register.json', RegisterHandler),
        (r'/api/update.json', UpdateHandler),
    ]
    app = tornado.web.Application(handlers=handlers)

    app.listen(8681, 'localhost')
    tornado.ioloop.IOLoop.instance().start()



import contextlib
import os
import traceback

import tornado.stack_context

from seesaw.event import Event


class Pipeline(object):
    '''The sequence of steps that complete a :class:`Task`.

    Your pipeline will probably be something like this:

    1. Request an assignment from the tracker.
    2. Run Wget to download the file.
    3. Upload the downloaded file with rsync.
    4. Tell the tracker that the assignment is done.
    '''
    def __init__(self, *tasks):
        self.cwd = os.getcwd()
        self.data_dir = os.path.join(self.cwd, "data")
        self.on_start_item = Event()
        self.on_complete_item = Event()
        self.on_fail_item = Event()
        self.on_cancel_item = Event()
        self.on_finish_item = Event()
        self.on_cleanup = Event()
        self.project = None

        self.items_in_pipeline = set()
        self.tasks = []
        for task in tasks:
            self.add_task(task)

    def add_task(self, task):
        task.on_complete_item += self._task_complete_item
        task.on_fail_item += self._task_fail_item
        self.tasks.append(task)

    def enqueue(self, item):
        self.items_in_pipeline.add(item)
        self.on_start_item(self, item)
        self._enqueue_with_except(self.tasks[0], item)

    def _enqueue_with_except(self, task, item):
        @contextlib.contextmanager
        def handle_item_exception(e_type, e_value, tb):
            item.log_output("Failed %s for %s\n" % (task, item.description()))
            item.log_output("%s\n" % traceback.format_exc())
            item.log_error(self, e_value)
            task.fail_item(item)

        with tornado.stack_context.NullContext():
            with tornado.stack_context.ExceptionStackContext(
                    handle_item_exception):
                task.enqueue(item)

    def _task_complete_item(self, task, item):
        task_index = self.tasks.index(task)
        if len(self.tasks) <= task_index + 1:
            self._complete_item(item)
        else:
            self._enqueue_with_except(self.tasks[task_index + 1], item)

    def _task_fail_item(self, task, item):
        self._fail_item(item)

    def _cancel_item(self, item):
        if item in self.items_in_pipeline:
            item.cancel()
            self.items_in_pipeline.remove(item)
            self.on_cancel_item(self, item)
            self.on_finish_item(self, item)
        else:
            # XXX: Reaching here indicates a programming problem.
            # Refactoring is required.
            item.log_output(
                'Warning: Ignoring extra cancel event.\n' +
                ''.join(traceback.format_stack()))

    def _complete_item(self, item):
        if item in self.items_in_pipeline:
            item.complete()
            self.items_in_pipeline.remove(item)
            self.on_complete_item(self, item)
            self.on_finish_item(self, item)
        else:
            # See comment above.
            item.log_output(
                'Warning: Ignoring extra complete event.\n' +
                ''.join(traceback.format_stack()))

    def _fail_item(self, item):
        if item in self.items_in_pipeline:
            item.fail()
            self.items_in_pipeline.remove(item)
            self.on_fail_item(self, item)
            self.on_finish_item(self, item)
        else:
            # See comment above.
            item.log_output(
                'Warning: Ignoring extra fail event.\n' +
                ''.join(traceback.format_stack()))

    def cancel_items(self):
        cancel_items = [item for item in self.items_in_pipeline
                        if item.may_be_canceled]

        for item in cancel_items:
            self._cancel_item(item)

    def ui_task_list(self):
        task_list = []
        for task in self.tasks:
            task.fill_ui_task_list(task_list)
        return task_list

    def __str__(self):
        return "Pipeline:\n -> " + ("\n -> ".join(map(str, self.tasks)))



'''Actor model.'''
# http://www.valuedlessons.com/2008/04/events-in-python.html


class Event(object):
    '''Lightweight event system.

    Example::

        my_event_system = Event()
        my_event_system = my_listener_callback_function
        my_event_system(my_event_data)
    '''
    def __init__(self):
        self.handlers = set()

    def handle(self, handler):
        self.handlers.add(handler)
        return self

    def unhandle(self, handler):
        try:
            self.handlers.remove(handler)
        except:
            raise ValueError("Handler is not handling this event, "
                             "so cannot unhandle it.")
        return self

    def fire(self, *args, **kargs):
        for handler in self.handlers:
            handler(*args, **kargs)

    def getHandlerCount(self):
        return len(self.handlers)

    __iadd__ = handle
    __isub__ = unhandle
    __call__ = fire
    __len__ = getHandlerCount



import base64
import re

import tornado


try:
    callable
except NameError:
    from seesaw.six import callable


class AuthenticationErrorHandler(tornado.web.RequestHandler):
    def initialize(self, realm="Restricted"):
        self.realm = realm

    def prepare(self):
        self.set_status(401)
        self.set_header('WWW-Authenticate', 'Basic realm=' + self.realm)
        self.finish()


class AuthenticatedApplication(tornado.web.Application):
    def __init__(self, *args, **kwargs):
        super(AuthenticatedApplication, self).__init__(*args, **kwargs)
        self.auth_enabled = kwargs.get("auth_enabled", True)
        self.auth_realm = kwargs.get("auth_realm", "Restricted")
        self.check_auth = kwargs.get("check_auth")
        self.skip_auth = [re.compile(pattern)
                          for pattern in kwargs.get("skip_auth", [])]

    def __call__(self, request):
        if not self.auth_enabled or (callable(self.auth_enabled) and
                                     not self.auth_enabled()):
            return super(AuthenticatedApplication, self).__call__(request)

        for pattern in self.skip_auth:
            if pattern.match(request.uri):
                return super(AuthenticatedApplication, self).__call__(request)

        auth_header = request.headers.get('Authorization')
        if auth_header and auth_header.startswith('Basic '):
            auth_decoded = base64.decodestring(auth_header[6:])
            username, password = auth_decoded.split(':', 2)
            request.basicauth_user, request.basicauth_pass = username, password

            if self.check_auth and \
                    self.check_auth(request, username, password):
                return super(AuthenticatedApplication, self).__call__(request)

        handler = AuthenticationErrorHandler(self, request,
                                             realm=self.auth_realm)
        handler._execute([])
        return handler



'''Running subprocesses asynchronously.'''
import fcntl
import os
import os.path
import subprocess
import functools
import datetime
import pty

import tornado.ioloop
from tornado.ioloop import IOLoop, PeriodicCallback

from seesaw.event import Event
from seesaw.task import Task
from seesaw.config import realize


class AsyncPopen(object):
    '''Asynchronous version of :class:`subprocess.Popen`.'''
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

        self.on_output = Event()
        self.on_end = Event()

    def run(self):
        self.ioloop = IOLoop.instance()
        (master_fd, slave_fd) = pty.openpty()

        # make stdout, stderr non-blocking
        fcntl.fcntl(master_fd, fcntl.F_SETFL,
                    fcntl.fcntl(master_fd, fcntl.F_GETFL) | os.O_NONBLOCK)

        self.master_fd = master_fd
        self.master = os.fdopen(master_fd)

        # listen to stdout, stderr
        self.ioloop.add_handler(master_fd, self._handle_subprocess_stdout,
                                self.ioloop.READ)

        slave = os.fdopen(slave_fd)
        self.kwargs["stdout"] = slave
        self.kwargs["stderr"] = slave
        self.kwargs["close_fds"] = True
        self.pipe = subprocess.Popen(*self.args, **self.kwargs)

        self.stdin = self.pipe.stdin

        # check for process exit
        self.wait_callback = PeriodicCallback(self._wait_for_end, 250)
        self.wait_callback.start()

    def _handle_subprocess_stdout(self, fd, events):
        if not self.master.closed and (events & IOLoop._EPOLLIN) != 0:
            data = self.master.read()
            self.on_output(data)

        self._wait_for_end(events)

    def _wait_for_end(self, events=0):
        self.pipe.poll()
        if self.pipe.returncode is not None or \
                (events & tornado.ioloop.IOLoop._EPOLLHUP) > 0:
            self.wait_callback.stop()
            self.master.close()
            self.ioloop.remove_handler(self.master_fd)
            self.on_end(self.pipe.returncode)


class ExternalProcess(Task):
    '''External subprocess runner.'''
    def __init__(self, name, args, max_tries=1, retry_delay=30,
                 accept_on_exit_code=None, retry_on_exit_code=None, env=None):
        Task.__init__(self, name)
        self.args = args
        self.max_tries = max_tries
        self.retry_delay = retry_delay
        if accept_on_exit_code is not None:
            self.accept_on_exit_code = accept_on_exit_code
        else:
            self.accept_on_exit_code = [0]
        self.retry_on_exit_code = retry_on_exit_code
        self.env = env or {}

        if 'PYTHONIOENCODING' not in self.env:
            self.env['PYTHONIOENCODING'] = 'utf8:replace'

    def enqueue(self, item):
        self.start_item(item)
        item.log_output("Starting %s for %s\n" % (self, item.description()))
        item["tries"] = 0
        item["ExternalProcess.stdin_write_error"] = False
        self.process(item)

    def stdin_data(self, item):
        return b""

    def process(self, item):
        with self.task_cwd():
            p = AsyncPopen(
                args=realize(self.args, item),
                env=realize(self.env, item),
                stdin=subprocess.PIPE,
                close_fds=True
            )

            p.on_output += functools.partial(self.on_subprocess_stdout, p,
                                             item)
            p.on_end += functools.partial(self.on_subprocess_end, item)

            p.run()

            try:
                p.stdin.write(self.stdin_data(item))
            except Exception as error:
                # FIXME: We need to properly propagate errors
                item.log_output("Error writing to process: %s" % str(error))
                item["ExternalProcess.stdin_write_error"] = True

            p.stdin.close()

    def on_subprocess_stdout(self, pipe, item, data):
        item.log_output(data, full_line=False)

    def on_subprocess_end(self, item, returncode):
        if returncode in self.accept_on_exit_code and \
                not item["ExternalProcess.stdin_write_error"]:
            self.handle_process_result(returncode, item)
        else:
            self.handle_process_error(returncode, item)

    def handle_process_result(self, exit_code, item):
        item.log_output("Finished %s for %s\n" % (self, item.description()))
        self.complete_item(item)

    def handle_process_error(self, exit_code, item):
        item["tries"] += 1

        item.log_output(
            "Process %s returned exit code %d for %s\n" %
            (self, exit_code, item.description())
        )
        item.log_error(self, exit_code)

        retry_acceptable = self.max_tries is None or \
            item["tries"] < self.max_tries
        exit_status_indicates_retry = self.retry_on_exit_code is None or \
            exit_code in self.retry_on_exit_code or \
            item["ExternalProcess.stdin_write_error"]

        if retry_acceptable and exit_status_indicates_retry:
            item.log_output(
                "Retrying %s for %s after %d seconds...\n" %
                (self, item.description(), self.retry_delay)
            )
            IOLoop.instance().add_timeout(
                datetime.timedelta(seconds=self.retry_delay),
                functools.partial(self.process, item)
            )

        else:
            item.log_output("Failed %s for %s\n" % (self, item.description()))
            self.fail_item(item)


class WgetDownload(ExternalProcess):
    '''Download with Wget process runner.'''
    def __init__(self, args, max_tries=1, accept_on_exit_code=None,
                 retry_on_exit_code=None, env=None, stdin_data_function=None):
        ExternalProcess.__init__(
            self, "WgetDownload",
            args=args, max_tries=max_tries,
            accept_on_exit_code=(accept_on_exit_code
                                 if accept_on_exit_code is not None else [0]),
            retry_on_exit_code=retry_on_exit_code,
            env=env)
        self.stdin_data_function = stdin_data_function

    def stdin_data(self, item):
        if self.stdin_data_function:
            return self.stdin_data_function(item)
        else:
            return b""


class RsyncUpload(ExternalProcess):
    '''Upload with Rsync process runner.'''
    def __init__(self, target, files, target_source_path="./", bwlimit="0",
                 max_tries=None, extra_args=None):
        args = [
            "rsync",
            "-avz",
            "--compress-level=9",
            "--timeout=300",
            "--contimeout=300",
            "--progress",
            "--bwlimit", bwlimit
        ]
        if extra_args is not None:
            args.extend(extra_args)
        args.extend([
            "--files-from=-",
            target_source_path,
            target
        ])
        ExternalProcess.__init__(self, "RsyncUpload",
                                 args=args,
                                 max_tries=max_tries)
        self.files = files
        self.target_source_path = target_source_path

    def stdin_data(self, item):
        return "".join(
            [
                "%s\n" % os.path.relpath(
                    realize(f, item),
                    realize(self.target_source_path, item)
                )
                for f in realize(self.files, item)
            ]).encode('utf-8')


class CurlUpload(ExternalProcess):
    '''Upload with Curl process runner.'''
    def __init__(self, target, filename, connect_timeout="60", speed_limit="1",
                 speed_time="900", max_tries=None):
        args = [
            "curl",
            "--fail",
            "--output", "/dev/null",
            "--connect-timeout", str(connect_timeout),
            "--speed-limit", str(speed_limit),  # minimum upload speed 1B/s
            # stop if speed < speed-limit for 900 seconds
            "--speed-time", str(speed_time),
            "--header", "X-Curl-Limits: inf,%s,%s" % (str(speed_limit),
                                                      str(speed_time)),
            "--write-out", "Upload server: %{url_effective}\\n",
            "--location",
            "--upload-file", filename,
            target
        ]
        ExternalProcess.__init__(self, "CurlUpload",
                                 args=args,
                                 max_tries=max_tries)



import logging
import tornado.ioloop
import unittest
import sys


class BaseTestCase(unittest.TestCase):
    def setUp(self):
        self.io_loop_error = None

        def periodic_callback_run_monkey_patch(self1):
            if not self1._running:
                return
            try:
                self1.callback()
            except:
                logging.exception('Periodic Callback')
                self.io_loop_error = True

            self1._schedule_next()

        tornado.ioloop.PeriodicCallback._run = \
            periodic_callback_run_monkey_patch

    def assertIOLoopOK(self):
        value = self.io_loop_error
        self.io_loop_error = None
        self.assertFalse(value)

    def tearDown(self):
        assert not self.io_loop_error



from seesaw.pipeline import Pipeline
from seesaw.runner import SimpleRunner
from seesaw.task import PrintItem, SimpleTask
from seesaw.test_base import BaseTestCase

class RunnerTest(BaseTestCase):
    def setUp(self):
        super(RunnerTest, self).setUp()

        self.cleanup_calls = 0

    def test_runner_does_pipeline_cleanup_before_shutdown(self):
        pipeline = Pipeline(PrintItem())
        runner = SimpleRunner(pipeline, max_items=1)

        def cleanup():
            self.cleanup_calls += 1

        pipeline.on_cleanup += cleanup
        runner.start()

        self.assertEqual(1, self.cleanup_calls)
        self.assertEqual(1, runner.item_count)



#!/usr/bin/env python
import sys

import seesaw

try:
    from setuptools import setup
    # hush pyflakes
    setup  # pylint: disable=pointless-statement
except ImportError:
    from distutils.core import setup


packages = [
    'seesaw',
    'seesaw.script',
]

package_dir = {
    'seesaw': 'seesaw',
}

package_data = {
    'seesaw': [
        'public/index.html',
        'public/*.js',
        'public/*.css',
        'templates/*.html'
    ]
}

if sys.version_info[0] == 3:
    scripts = [
        'run-pipeline3',
        'run-warrior3',
    ]
else:
    scripts = [
        'run-pipeline',
        'run-warrior',
        'run-pipeline2',
        'run-warrior2',
    ]

requires = [
    'Tornado>=2.3',
    'sockjs-tornado',
]

if sys.version_info < (2, 7):
    requires.append('ordereddict')

setup(
    name='seesaw',
    version=seesaw.__version__,
    maintainer='ArchiveTeam',
    maintainer_email='warrior@archiveteam.org',
    description='ArchiveTeam seesaw kit',
    long_description=open('README.md', 'r').read(),
    url='http://www.archiveteam.org/',
    packages=packages,
    package_dir=package_dir,
    package_data=package_data,
    include_package_data=True,
    scripts=scripts,
    install_requires=requires,
)



# encoding=utf-8
from __future__ import unicode_literals

import datetime
import os
import os.path
import shutil
import json

from seesaw.project import *
from seesaw.item import *
from seesaw.task import *
from seesaw.pipeline import *
from seesaw.externalprocess import *
from seesaw.tracker import *

project = Project(
  title = "Example project",
  project_html = """
    <img class="project-logo" alt="Project logo" src="http://archive.org/images/glogo.png" height="50px" />
    <h2>Example project <span class="links"><a href="http://example.com/">Example website</a> &middot; <a href="http://example.heroku.com/">Leaderboard</a></span></h2>
    <p>This is an example project. Under a logo and title there's some room for extra information.</p>
  """,
  utc_deadline = datetime.datetime(2013,1,1, 12,0,0)
)

class CustomTask(SimpleTask):
    def __init__(self):
        SimpleTask.__init__(self, 'CustomTask')

    def process(self, item):
        item.log_output('ÐÐ Hello !')

        # Test binary output
        item.log_output('ðbßf'.encode('utf-8'))
        item.log_output(b'\xff\xff\xff')


pipeline = Pipeline(
  SetItemKey("item_name", "1083030"),
  PrintItem(),
  CustomTask(),
  ExternalProcess("Echo", [ "echo", "1234" ]),
  ExternalProcess("sleep", [ "sleep", "5" ]),
  ExternalProcess("pwd", [ "pwd" ]),
  PrintItem()
)




import distutils.version
import seesaw
import unittest


class InitTest(unittest.TestCase):
    def test_valid_version(self):
        '''It should not raise ValueError.'''
        distutils.version.StrictVersion(seesaw.__version__)

    def test_valid_build_number(self):
        '''It should match the version string.'''
        version = distutils.version.StrictVersion(seesaw.__version__)
        major_ver, minor_ver, patch_ver = version.version
        major_build_ver = (seesaw.__build__ & 0xff0000) >> 16
        minor_build_ver = (seesaw.__build__ & 0xff00) >> 8
        patch_build_ver = seesaw.__build__ & 0xff

        self.assertEqual(major_ver, major_build_ver)
        self.assertEqual(minor_ver, minor_build_ver)
        self.assertEqual(patch_ver, patch_build_ver)



'''Managing work units.'''
import os
import os.path
import shutil
import traceback

from seesaw.event import Event
import seesaw.six


class Item(object):
    '''A thing, or work unit, that needs to be downloaded.

    It has properties that are filled by the :class:`Task`.

    An Item behaves like a mutable mapping.

    .. note::
        State belonging to a item should be stored on the actual item
        itself. That is, do not store variables onto a :class:`Task` unless
        you know what you are doing.
    '''
    def __init__(self, pipeline, item_id, item_number, properties=None,
                 keep_data=False, prepare_data_directory=True):
        self.pipeline = pipeline
        self.item_id = item_id
        self.item_number = item_number
        self.properties = properties or {}
        self.keep_data = keep_data

        self.may_be_canceled = False
        self.canceled = False
        self.completed = False
        self.failed = False
        self.finished = False
        self._errors = []
        self._last_output = ""
        self.task_status = {}

        self.on_output = Event()
        self.on_error = Event()
        self.on_task_status = Event()
        self.on_property = Event()
        self.on_cancel = Event()
        self.on_complete = Event()
        self.on_fail = Event()
        self.on_finish = Event()

        if prepare_data_directory:
            self.prepare_data_directory()

    def prepare_data_directory(self):
        dirname = os.path.join(self.pipeline.data_dir, self.item_id)
        self["data_dir"] = dirname
        if os.path.isdir(dirname):
            shutil.rmtree(dirname)
        os.makedirs(dirname)

    def clear_data_directory(self):
        if not self.keep_data:
            dirname = self["data_dir"]
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

    def log_output(self, data, full_line=True):
        if isinstance(data, seesaw.six.binary_type):
            try:
                data = data.decode('utf8', 'replace')
            except UnicodeError:
                data = data.decode('ascii', 'replace')

        if full_line and len(data) > 0:
            if data[0] != "\n" and len(self._last_output) > 0 and \
                    self._last_output[-1] != "\n":
                data = "\n" + data
            if data[-1] != "\n":
                data = data + "\n"
        self._last_output = data
        self.on_output(self, data)

    def log_error(self, task, *args):
        self._errors.append((task, args))
        self.on_error(self, task, *args)

    def set_task_status(self, task, status):
        if task in self.task_status:
            old_status = self.task_status[task]
        else:
            old_status = None
        if status != old_status:
            self.task_status[task] = status
            self.on_task_status(self, task, status, old_status)

    def cancel(self):
        assert not self.canceled
        self.clear_data_directory()
        self.canceled = True
        self.finished = True
        self.on_cancel(self)
        self.on_finish(self)

    def complete(self):
        assert not self.completed
        self.clear_data_directory()
        self.completed = True
        self.finished = True
        self.on_complete(self)
        self.on_finish(self)

    def fail(self):
        assert not self.failed
        self.clear_data_directory()
        self.failed = True
        self.finished = True
        self.on_fail(self)
        self.on_finish(self)

    def description(self):
        return "Item %s" % (self.properties["item_name"]
                            if "item_name" in self.properties else "")

    def get(self, key):
        return self.properties.get(key)

    def __contains__(self, key):
        return key in self.properties

    def __getitem__(self, key):
        return self.properties[key]

    def __setitem__(self, key, value):
        old_value = self.properties[key] if key in self.properties else None
        self.properties[key] = value
        if old_value != value:
            self.on_property(self, key, value, old_value)

    def __delitem__(self, key):
        old_value = self.properties[key] if key in self.properties else None
        del self.properties[key]
        if old_value:
            self.on_property(self, key, None, old_value)

    def __str__(self):
        s = "Item " + ("FAILED " if self.failed else "") + str(self.properties)
        for err in self._errors:
            for e in err[1]:
                # TODO this isn't how exceptions work?
                if isinstance(e, Exception):
                    s += "%s\n" % traceback.format_exception(Exception, e,
                                                             None)
                else:
                    s += "%s\n" % str(e)
            s += "\n  " + str(err)
        return s

    class TaskStatus(object):
        running = "running"
        completed = "completed"
        failed = "failed"


class ItemValue(object):
    '''Get an item's value during :func:`realize`.'''
    def __init__(self, key):
        self.key = key

    def realize(self, item):
        return item[self.key]

    def fill(self, item, value):
        if isinstance(self, ItemValue):
            item[self.key] = value
        elif self is None:
            pass
        else:
            raise Exception("Attempting to fill " + str(type(self)))

    def __str__(self):
        return "<" + self.key + ">"


class ItemInterpolation(object):
    '''Formats a string using the percent operator during :func:`realize`.'''
    def __init__(self, s):
        self.s = s

    def realize(self, item):
        return self.s % item

    def __str__(self):
        return "<'" + self.s + "'>"



'''The warrior web interface.'''
import json
import os
import os.path
import random
import re
import time

from sockjs.tornado import SockJSConnection, SockJSRouter
from tornado import web, ioloop

from seesaw.config import realize
from seesaw.web_util import AuthenticatedApplication


PUBLIC_PATH = os.path.abspath(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "public"))
TEMPLATES_PATH = os.path.abspath(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates"))


class IndexHandler(web.RequestHandler):
    '''Shows the index.html.'''
    def get(self):
        self.render(os.path.join(PUBLIC_PATH, "index.html"),
                    timestamp=time.time())


class ItemMonitor(object):
    '''Pushes item states and information to the client.'''
    def __init__(self, item):
        self.pipeline = item.pipeline
        self.item = item

        item.on_output += self.handle_item_output
        item.on_task_status += self.handle_item_task_status
        item.on_property += self.handle_item_property
        item.on_complete += self.handle_item_complete
        item.on_fail += self.handle_item_fail
        item.on_cancel += self.handle_item_cancel

        self.collected_data = []

        SeesawConnection.broadcast(
            "pipeline.start_item",
            {"pipeline_id": id(self.pipeline),
                "item": self.item_for_broadcast()}
            )

    def item_for_broadcast(self):
        item = self.item

        tasks = []
        for (task, task_name) in self.pipeline.ui_task_list():
            tasks.append({
                "id": id(task),
                "name": task_name,
                "status": (item.task_status[task]
                           if task in item.task_status else None)
            })

        if self.pipeline.project:
            project_name = self.pipeline.project.title
        else:
            project_name = None
        item_data = {
            "id": item.item_id,
            "name": ("Item %s" % item["item_name"]
                     if "item_name" in item else "New item"),
            "number": item.item_number,
            "status": self.item_status(),
            "tasks": tasks,
            "output": "".join(self.collected_data),
            "project": project_name
        }

        return item_data

    def item_status(self):
        if self.item.completed:
            return "completed"
        elif self.item.failed:
            return "failed"
        elif self.item.canceled:
            return "canceled"
        else:
            return "running"

    def handle_item_output(self, item, data):
        self.collected_data.append(data)
        SeesawConnection.broadcast(
            "item.output", {"item_id": item.item_id, "data": data})

    def handle_item_task_status(self, item, task, new_status, old_status):
        SeesawConnection.broadcast(
            "item.task_status",
            {
                "item_id": item.item_id,
                "task_id": id(task),
                "new_status": new_status,
                "old_status": old_status
            }
        )

    def handle_item_property(self, item, key, new_value, old_value):
        if key == "item_name":
            SeesawConnection.broadcast(
                "item.update_name",
                {"item_id": item.item_id, "new_name": "Item %s" % new_value})

    def handle_item_complete(self, item):
        SeesawConnection.broadcast("item.complete", {"item_id": item.item_id})

    def handle_item_fail(self, item):
        SeesawConnection.broadcast("item.fail", {"item_id": item.item_id})

    def handle_item_cancel(self, item):
        SeesawConnection.broadcast("item.cancel", {"item_id": item.item_id})


class ApiHandler(web.RequestHandler):
    '''Processes API requests.'''
    def initialize(self, warrior=None, runner=None):
        self.warrior = warrior
        self.runner = runner

    def get_template_path(self):
        return TEMPLATES_PATH

    def post(self, command):
        if command == "stop":
            if self.warrior:
                self.warrior.stop_gracefully()
            else:
                self.runner.stop_gracefully()
            self.write("OK")
        elif command == "stop_now":
            if self.warrior:
                self.warrior.forced_stop()
            else:
                self.runner.forced_stop()
            self.write("OK")
        elif command == "keep_running":
            if self.warrior:
                self.warrior.keep_running()
            else:
                self.runner.keep_running()
            self.write("OK")
        elif command == "select-project":
            self.warrior.config_manager.set_value(
                "selected_project", self.get_argument("project_name"))
            self.warrior.select_project(self.get_argument("project_name"))
            self.write("OK")
        elif command == "deselect-project":
            self.warrior.config_manager.set_value("selected_project", "none")
            self.warrior.select_project(None)
            self.write("OK")
        elif command == "settings":
            posted_values = {}
            for (name, value) in self.request.arguments.items():
                value[0] = value[0].decode('utf8', 'replace')

                if not self.warrior.config_manager.set_value(name, value[0]):
                    posted_values[name] = value[0]
            if self.warrior.config_manager.all_valid():
                self.warrior.fire_status()
            self.render("settings.html", warrior=self.warrior,
                        posted_values=posted_values)

    def get(self, command):
        if command == "all-projects":
            self.render("all-projects.html", warrior=self.warrior,
                        realize=realize)
        elif command == "settings":
            self.render("settings.html", warrior=self.warrior,
                        posted_values={})
        elif command == "help":
            self.render("help.html", warrior=self.warrior)


class SeesawConnection(SockJSConnection):
    '''A WebSocket server that communicates the state of the warrior.'''
    instance_id = ("%d-%f" % (os.getpid(), random.random()))

    clients = set()
    item_monitors = dict()

    warrior = None
    project = None
    runner = None

    def emit(self, event_name, message):
        '''tornadoio to sockjs adapter.'''
        self.send(json.dumps({'event_name': event_name, 'message': message}))

    def on_open(self, info):
        self.clients.add(self)
        self.emit("instance_id", self.instance_id)

        items = []
        for item_monitor in self.item_monitors.values():
            items.append(item_monitor.item_for_broadcast())

        if self.project:
            self.emit("project.refresh", {
                "project": self.project.data_for_json(),
                "status": ("stopping"
                           if self.runner.should_stop() else "running"),
                "items": items
            })
        else:
            self.emit("project.refresh", None)

        if self.warrior:
            self.emit("warrior.projects_loaded", {
                "projects": self.warrior.projects
            })
            self.emit("warrior.status",
                      {"status": self.warrior.warrior_status()})

    @classmethod
    def broadcast_bandwidth(cls):
        if cls.warrior:
            bw_stats = cls.warrior.bandwidth_stats()
            if bw_stats:
                cls.broadcast("bandwidth", bw_stats)

    @classmethod
    def handle_warrior_status(cls, warrior, new_status):
        cls.broadcast("warrior.status", {"status": new_status})

    @classmethod
    def handle_projects_loaded(cls, warrior, projects):
        cls.broadcast_projects()

    @classmethod
    def broadcast_projects(cls):
        cls.broadcast("warrior.projects_loaded", {
            "projects": cls.warrior.projects
        })

    @classmethod
    def handle_project_selected(cls, warrior, project):
        cls.broadcast("warrior.project_selected", {"project": project})

    @classmethod
    def handle_project_installing(cls, warrior, project):
        cls.broadcast("warrior.project_installing", {"project": project})

    @classmethod
    def handle_project_installed(cls, warrior, project, output):
        output = re.sub("\r\n?", "\n", output)
        cls.broadcast("warrior.project_installed",
                      {"project": project, "output": output})

    @classmethod
    def handle_project_installation_failed(cls, warrior, project, output):
        output = re.sub("\r\n?", "\n", output)
        cls.broadcast("warrior.project_installation_failed",
                      {"project": project, "output": output})

    @classmethod
    def handle_project_refresh(cls, warrior, project, runner):
        cls.project = project
        cls.runner = runner
        cls.broadcast_project_refresh()

    @classmethod
    def broadcast_project_refresh(cls):
        if cls.project:
            cls.broadcast("project.refresh", {
                "project": cls.project.data_for_json(),
                "status": ("stopping"
                           if cls.runner.should_stop() else "running"),
                "items": []
            })
        else:
            cls.broadcast("project.refresh", None)

    @classmethod
    def handle_runner_status(cls, runner, status):
        cls.broadcast("runner.status", {
            "status": ("stopping" if runner.should_stop() else "running")
        })

    @classmethod
    def handle_start_item(cls, runner, pipeline, item):
        cls.item_monitors[item] = ItemMonitor(item)

    @classmethod
    def handle_finish_item(cls, runner, pipeline, item):
        del cls.item_monitors[item]

    @classmethod
    def broadcast(cls, event, message):
        for client in cls.clients:
            if message:
                message["session_id"] = client.session.session_id
            client.emit(event, message)

    def on_message(self, message):
        pass

    def on_close(self):
        self.clients.remove(self)


def start_runner_server(project, runner, bind_address="", port_number=8001,
                        http_username=None, http_password=None):
    '''Starts a web interface for a manually run pipeline.

    Unlike :func:`start_warrior_server`, this UI does not contain an
    configuration or project management panel.
    '''
    if bind_address == "0.0.0.0":
        bind_address = ""

    SeesawConnection.project = project
    SeesawConnection.runner = runner

    runner.on_pipeline_start_item += SeesawConnection.handle_start_item
    runner.on_pipeline_finish_item += SeesawConnection.handle_finish_item
    runner.on_status += SeesawConnection.handle_runner_status

    router = SockJSRouter(SeesawConnection)

    application = AuthenticatedApplication(
        router.apply_routes([
            (r"/(.*\.(html|css|js|swf|png|ico))$",
                web.StaticFileHandler, {"path": PUBLIC_PATH}),
            ("/", IndexHandler),
            ("/api/(.+)$", ApiHandler, {"runner": runner})]),
        #  flash_policy_port = 843,
        #  flash_policy_file=os.path.join(PUBLIC_PATH, "flashpolicy.xml"),
        socket_io_address=bind_address,
        socket_io_port=port_number,

        # settings for AuthenticatedApplication
        auth_enabled=(http_password or "").strip() != "",
        check_auth=lambda r, username, password:
            (
                password == http_password and
                (http_username or "").strip() in ["", username]
            ),
        auth_realm="ArchiveTeam Warrior",
        skip_auth=[r"^/socket\.io/1/websocket/[a-z0-9]+$"]
    )

    application.listen(port_number)


def start_warrior_server(warrior, bind_address="", port_number=8001,
                         http_username=None, http_password=None):
    '''Starts the warrior web interface.'''
    SeesawConnection.warrior = warrior

    warrior.on_projects_loaded += SeesawConnection.handle_projects_loaded
    warrior.on_project_refresh += SeesawConnection.handle_project_refresh
    warrior.on_project_installing += SeesawConnection.handle_project_installing
    warrior.on_project_installed += SeesawConnection.handle_project_installed
    warrior.on_project_installation_failed += \
        SeesawConnection.handle_project_installation_failed
    warrior.on_project_selected += SeesawConnection.handle_project_selected
    warrior.on_status += SeesawConnection.handle_warrior_status
    warrior.runner.on_pipeline_start_item += SeesawConnection.handle_start_item
    warrior.runner.on_pipeline_finish_item += \
        SeesawConnection.handle_finish_item
    warrior.runner.on_status += SeesawConnection.handle_runner_status

    if not http_username:
        http_username = warrior.http_username
    if not http_password:
        http_password = warrior.http_password

    ioloop.PeriodicCallback(SeesawConnection.broadcast_bandwidth, 1000).start()

    router = SockJSRouter(SeesawConnection)

    application = AuthenticatedApplication(
        router.apply_routes([
            (r"/(.*\.(html|css|js|swf|png|ico))$",
                web.StaticFileHandler, {"path": PUBLIC_PATH}),
            ("/", IndexHandler),
            ("/api/(.+)$", ApiHandler, {"warrior": warrior})]),
        #   flash_policy_port = 843,
        #   flash_policy_file = os.path.join(PUBLIC_PATH, "flashpolicy.xml"),
        socket_io_address=bind_address,
        socket_io_port=port_number,

        # settings for AuthenticatedApplication
        auth_enabled=lambda: (realize(http_password) or "").strip() != "",
        check_auth=lambda r, username, password:
            (
                password == realize(http_password) and
                (realize(http_username) or "").strip() in ["", username]
            ),
        auth_realm="ArchiveTeam Warrior",
        skip_auth=[tornado_url[0] for tornado_url in router.urls]
    )

    application.listen(port_number)



#!/usr/bin/python3
'''Runs the complete warrior system and the Mock HQ.'''
import atexit
import copy
import os.path
import shutil
import subprocess
import tempfile


def main():
    temp_dir = tempfile.mkdtemp(prefix='warrior-tmp')
    this_dir = os.path.dirname(__file__)
    root_dir = os.path.join(this_dir, '..')
    env = copy.copy(os.environ)
    env['PYTHONPATH'] = root_dir

    def cleanup():
        shutil.rmtree(temp_dir)

    atexit.register(cleanup)

    web_process = subprocess.Popen([
        os.environ.get('PYTHON_BIN', 'python3'), os.path.join(this_dir, 'hq_mock.py'),
        ],
    )

    subprocess.check_call([
        os.path.join(root_dir, './run-warrior3'),
        '--projects-dir', temp_dir,
        '--data-dir', temp_dir,
        '--address', 'localhost',
        '--port', '8688',
        '--warrior-hq', 'http://localhost:8681/',
        ],
        env=env
    )

    web_process.terminate()

if __name__ == '__main__':
    main()
